import os

from clang.cindex import Config, TranslationUnit

if "CLANG_LIBRARY_PATH" in os.environ:
    Config.set_library_path(os.environ["CLANG_LIBRARY_PATH"])

import unittest
from pathlib import Path


class TestCodeCompletion(unittest.TestCase):
    def check_completion_results(self, cr, expected):
        self.assertIsNotNone(cr)
        self.assertEqual(len(cr.diagnostics), 0)

        completions = [str(c) for c in cr.results]

        for c in expected:
            self.assertIn(c, completions)

    def test_code_complete(self):
        files = [
            (
                "fake.c",
                """
/// Aaa.
int test1;

/// Bbb.
void test2(void);

void f() {

}
""",
            )
        ]

        tu = TranslationUnit.from_source(
            "fake.c",
            ["-std=c99"],
            unsaved_files=files,
            options=TranslationUnit.PARSE_INCLUDE_BRIEF_COMMENTS_IN_CODE_COMPLETION,
        )

        cr = tu.codeComplete(
            "fake.c", 9, 1, unsaved_files=files, include_brief_comments=True
        )

        expected = [
            "{'int', ResultType} | {'test1', TypedText} || Priority: 50 || Availability: Available || Brief comment: Aaa.",
            "{'void', ResultType} | {'test2', TypedText} | {'(', LeftParen} | {')', RightParen} || Priority: 50 || Availability: Available || Brief comment: Bbb.",
            "{'return', TypedText} | {';', SemiColon} || Priority: 40 || Availability: Available || Brief comment: ",
        ]
        self.check_completion_results(cr, expected)

    def test_code_complete_pathlike(self):
        files = [
            (
                Path("fake.c"),
                """
/// Aaa.
int test1;

/// Bbb.
void test2(void);

void f() {

}
""",
            )
        ]

        tu = TranslationUnit.from_source(
            Path("fake.c"),
            ["-std=c99"],
            unsaved_files=files,
            options=TranslationUnit.PARSE_INCLUDE_BRIEF_COMMENTS_IN_CODE_COMPLETION,
        )

        cr = tu.codeComplete(
            Path("fake.c"),
            9,
            1,
            unsaved_files=files,
            include_brief_comments=True,
        )

        expected = [
            "{'int', ResultType} | {'test1', TypedText} || Priority: 50 || Availability: Available || Brief comment: Aaa.",
            "{'void', ResultType} | {'test2', TypedText} | {'(', LeftParen} | {')', RightParen} || Priority: 50 || Availability: Available || Brief comment: Bbb.",
            "{'return', TypedText} | {';', SemiColon} || Priority: 40 || Availability: Available || Brief comment: ",
        ]
        self.check_completion_results(cr, expected)

    def test_code_complete_availability(self):
        files = [
            (
                "fake.cpp",
                """
class P {
protected:
  int member;
};

class Q : public P {
public:
  using P::member;
};

void f(P x, Q y) {
  x.; // member is inaccessible
  y.; // member is accessible
}
""",
            )
        ]

        tu = TranslationUnit.from_source(
            "fake.cpp", ["-std=c++98"], unsaved_files=files
        )

        cr = tu.codeComplete("fake.cpp", 12, 5, unsaved_files=files)

        expected = [
            "{'const', TypedText} || Priority: 50 || Availability: Available || Brief comment: ",
            "{'volatile', TypedText} || Priority: 50 || Availability: Available || Brief comment: ",
            "{'operator', TypedText} || Priority: 40 || Availability: Available || Brief comment: ",
            "{'P', TypedText} || Priority: 50 || Availability: Available || Brief comment: ",
            "{'Q', TypedText} || Priority: 50 || Availability: Available || Brief comment: ",
        ]
        self.check_completion_results(cr, expected)

        cr = tu.codeComplete("fake.cpp", 13, 5, unsaved_files=files)
        expected = [
            "{'P', TypedText} | {'::', Text} || Priority: 75 || Availability: Available || Brief comment: ",
            "{'P &', ResultType} | {'operator=', TypedText} | {'(', LeftParen} | {'const P &', Placeholder} | {')', RightParen} || Priority: 79 || Availability: Available || Brief comment: ",
            "{'int', ResultType} | {'member', TypedText} || Priority: 35 || Availability: NotAccessible || Brief comment: ",
            "{'void', ResultType} | {'~P', TypedText} | {'(', LeftParen} | {')', RightParen} || Priority: 79 || Availability: Available || Brief comment: ",
        ]
        self.check_completion_results(cr, expected)



import os

from clang.cindex import Config, Index, TranslationUnit

if "CLANG_LIBRARY_PATH" in os.environ:
    Config.set_library_path(os.environ["CLANG_LIBRARY_PATH"])

import unittest

INPUTS_DIR = os.path.join(os.path.dirname(__file__), "INPUTS")


class TestIndex(unittest.TestCase):
    def test_create(self):
        Index.create()

    # FIXME: test Index.read

    def test_parse(self):
        index = Index.create()
        self.assertIsInstance(index, Index)
        tu = index.parse(os.path.join(INPUTS_DIR, "hello.cpp"))
        self.assertIsInstance(tu, TranslationUnit)
        tu = index.parse(None, ["-c", os.path.join(INPUTS_DIR, "hello.cpp")])
        self.assertIsInstance(tu, TranslationUnit)



import os

from clang.cindex import Config, Diagnostic

if "CLANG_LIBRARY_PATH" in os.environ:
    Config.set_library_path(os.environ["CLANG_LIBRARY_PATH"])

import unittest

from .util import get_tu

# FIXME: We need support for invalid translation units to test better.


class TestDiagnostics(unittest.TestCase):
    def test_diagnostic_warning(self):
        tu = get_tu("int f0() {}\n")
        self.assertEqual(len(tu.diagnostics), 1)
        self.assertEqual(tu.diagnostics[0].severity, Diagnostic.Warning)
        self.assertEqual(tu.diagnostics[0].location.line, 1)
        self.assertEqual(tu.diagnostics[0].location.column, 11)
        self.assertEqual(
            tu.diagnostics[0].spelling, "non-void function does not return a value"
        )

    def test_diagnostic_note(self):
        # FIXME: We aren't getting notes here for some reason.
        tu = get_tu("#define A x\nvoid *A = 1;\n")
        self.assertEqual(len(tu.diagnostics), 1)
        self.assertEqual(tu.diagnostics[0].severity, Diagnostic.Error)
        self.assertEqual(tu.diagnostics[0].location.line, 2)
        self.assertEqual(tu.diagnostics[0].location.column, 7)
        self.assertIn("incompatible", tu.diagnostics[0].spelling)

    #       self.assertEqual(tu.diagnostics[1].severity, Diagnostic.Note)
    #       self.assertEqual(tu.diagnostics[1].location.line, 1)
    #       self.assertEqual(tu.diagnostics[1].location.column, 11)
    #       self.assertEqual(tu.diagnostics[1].spelling, 'instantiated from')

    def test_diagnostic_fixit(self):
        tu = get_tu("struct { int f0; } x = { f0 : 1 };")
        self.assertEqual(len(tu.diagnostics), 1)
        self.assertEqual(tu.diagnostics[0].severity, Diagnostic.Warning)
        self.assertEqual(tu.diagnostics[0].location.line, 1)
        self.assertEqual(tu.diagnostics[0].location.column, 26)
        self.assertRegex(tu.diagnostics[0].spelling, "use of GNU old-style.*")
        self.assertEqual(len(tu.diagnostics[0].fixits), 1)
        with self.assertRaises(IndexError):
            tu.diagnostics[0].fixits[1]
        self.assertEqual(tu.diagnostics[0].fixits[0].range.start.line, 1)
        self.assertEqual(tu.diagnostics[0].fixits[0].range.start.column, 26)
        self.assertEqual(tu.diagnostics[0].fixits[0].range.end.line, 1)
        self.assertEqual(tu.diagnostics[0].fixits[0].range.end.column, 30)
        self.assertEqual(tu.diagnostics[0].fixits[0].value, ".f0 = ")

    def test_diagnostic_range(self):
        tu = get_tu('void f() { int i = "a"; }')
        self.assertEqual(len(tu.diagnostics), 1)
        self.assertEqual(tu.diagnostics[0].severity, Diagnostic.Error)
        self.assertEqual(tu.diagnostics[0].location.line, 1)
        self.assertEqual(tu.diagnostics[0].location.column, 16)
        self.assertRegex(tu.diagnostics[0].spelling, "incompatible pointer to.*")
        self.assertEqual(len(tu.diagnostics[0].fixits), 0)
        self.assertEqual(len(tu.diagnostics[0].ranges), 1)
        self.assertEqual(tu.diagnostics[0].ranges[0].start.line, 1)
        self.assertEqual(tu.diagnostics[0].ranges[0].start.column, 20)
        self.assertEqual(tu.diagnostics[0].ranges[0].end.line, 1)
        self.assertEqual(tu.diagnostics[0].ranges[0].end.column, 23)
        with self.assertRaises(IndexError):
            tu.diagnostics[0].ranges[1].start.line

    def test_diagnostic_category(self):
        """Ensure that category properties work."""
        tu = get_tu("int f(int i) { return 7; }", all_warnings=True)
        self.assertEqual(len(tu.diagnostics), 1)
        d = tu.diagnostics[0]

        self.assertEqual(d.severity, Diagnostic.Warning)
        self.assertEqual(d.location.line, 1)
        self.assertEqual(d.location.column, 11)

        self.assertEqual(d.category_number, 3)
        self.assertEqual(d.category_name, "Semantic Issue")

    def test_diagnostic_option(self):
        """Ensure that category option properties work."""
        tu = get_tu("int f(int i) { return 7; }", all_warnings=True)
        self.assertEqual(len(tu.diagnostics), 1)
        d = tu.diagnostics[0]

        self.assertEqual(d.option, "-Wunused-parameter")
        self.assertEqual(d.disable_option, "-Wno-unused-parameter")

    def test_diagnostic_children(self):
        tu = get_tu("void f(int x) {} void g() { f(); }")
        self.assertEqual(len(tu.diagnostics), 1)
        d = tu.diagnostics[0]

        children = d.children
        self.assertEqual(len(children), 1)
        with self.assertRaises(IndexError):
            children[1]
        self.assertEqual(children[0].severity, Diagnostic.Note)
        self.assertRegex(children[0].spelling, ".*declared here")
        self.assertEqual(children[0].location.line, 1)
        self.assertEqual(children[0].location.column, 6)

    def test_diagnostic_string_repr(self):
        tu = get_tu("struct MissingSemicolon{}")
        self.assertEqual(len(tu.diagnostics), 1)
        d = tu.diagnostics[0]

        self.assertEqual(
            repr(d),
            "<Diagnostic severity 3, location <SourceLocation file 't.c', line 1, column 26>, spelling \"expected ';' after struct\">",
        )

    def test_diagnostic_string_format(self):
        tu = get_tu("struct MissingSemicolon{}")
        self.assertEqual(len(tu.diagnostics), 1)
        d = tu.diagnostics[0]

        self.assertEqual(str(d), "t.c:1:26: error: expected ';' after struct")
        self.assertEqual(
            d.format(0b111111),
            "t.c:1:26: error: expected ';' after struct [2, Parse Issue]",
        )
        with self.assertRaises(ValueError):
            d.format(0b1000000)



import os

from clang.cindex import (
    Config,
    CursorKind,
    PrintingPolicy,
    PrintingPolicyProperty,
    RefQualifierKind,
    TranslationUnit,
    TypeKind,
)

if "CLANG_LIBRARY_PATH" in os.environ:
    Config.set_library_path(os.environ["CLANG_LIBRARY_PATH"])

import gc
import unittest

from .util import get_cursor, get_cursors, get_tu

STRUCT_INPUT = """\

typedef int I;

struct teststruct {
  int a;
  I b;
  long c;
  unsigned long d;
  signed long e;
  const int f;
  int *g;
  int ***h;
};

"""


CONSTARRAY_INPUT = """
struct teststruct {
  void *A[2];
};
"""


class TestType(unittest.TestCase):
    def test_a_struct(self):
        tu = get_tu(STRUCT_INPUT)

        teststruct = get_cursor(tu, "teststruct")
        self.assertIsNotNone(teststruct, "Could not find teststruct.")
        fields = list(teststruct.get_children())

        self.assertEqual(fields[0].kind, CursorKind.FIELD_DECL)
        self.assertIsNotNone(fields[0].translation_unit)
        self.assertEqual(fields[0].spelling, "a")
        self.assertFalse(fields[0].type.is_const_qualified())
        self.assertEqual(fields[0].type.kind, TypeKind.INT)
        self.assertEqual(fields[0].type.get_canonical().kind, TypeKind.INT)
        self.assertEqual(fields[0].type.get_typedef_name(), "")

        self.assertEqual(fields[1].kind, CursorKind.FIELD_DECL)
        self.assertIsNotNone(fields[1].translation_unit)
        self.assertEqual(fields[1].spelling, "b")
        self.assertFalse(fields[1].type.is_const_qualified())
        self.assertEqual(fields[1].type.kind, TypeKind.ELABORATED)
        self.assertEqual(fields[1].type.get_canonical().kind, TypeKind.INT)
        self.assertEqual(fields[1].type.get_declaration().spelling, "I")
        self.assertEqual(fields[1].type.get_typedef_name(), "I")

        self.assertEqual(fields[2].kind, CursorKind.FIELD_DECL)
        self.assertIsNotNone(fields[2].translation_unit)
        self.assertEqual(fields[2].spelling, "c")
        self.assertFalse(fields[2].type.is_const_qualified())
        self.assertEqual(fields[2].type.kind, TypeKind.LONG)
        self.assertEqual(fields[2].type.get_canonical().kind, TypeKind.LONG)
        self.assertEqual(fields[2].type.get_typedef_name(), "")

        self.assertEqual(fields[3].kind, CursorKind.FIELD_DECL)
        self.assertIsNotNone(fields[3].translation_unit)
        self.assertEqual(fields[3].spelling, "d")
        self.assertFalse(fields[3].type.is_const_qualified())
        self.assertEqual(fields[3].type.kind, TypeKind.ULONG)
        self.assertEqual(fields[3].type.get_canonical().kind, TypeKind.ULONG)
        self.assertEqual(fields[3].type.get_typedef_name(), "")

        self.assertEqual(fields[4].kind, CursorKind.FIELD_DECL)
        self.assertIsNotNone(fields[4].translation_unit)
        self.assertEqual(fields[4].spelling, "e")
        self.assertFalse(fields[4].type.is_const_qualified())
        self.assertEqual(fields[4].type.kind, TypeKind.LONG)
        self.assertEqual(fields[4].type.get_canonical().kind, TypeKind.LONG)
        self.assertEqual(fields[4].type.get_typedef_name(), "")

        self.assertEqual(fields[5].kind, CursorKind.FIELD_DECL)
        self.assertIsNotNone(fields[5].translation_unit)
        self.assertEqual(fields[5].spelling, "f")
        self.assertTrue(fields[5].type.is_const_qualified())
        self.assertEqual(fields[5].type.kind, TypeKind.INT)
        self.assertEqual(fields[5].type.get_canonical().kind, TypeKind.INT)
        self.assertEqual(fields[5].type.get_typedef_name(), "")

        self.assertEqual(fields[6].kind, CursorKind.FIELD_DECL)
        self.assertIsNotNone(fields[6].translation_unit)
        self.assertEqual(fields[6].spelling, "g")
        self.assertFalse(fields[6].type.is_const_qualified())
        self.assertEqual(fields[6].type.kind, TypeKind.POINTER)
        self.assertEqual(fields[6].type.get_pointee().kind, TypeKind.INT)
        self.assertEqual(fields[6].type.get_typedef_name(), "")

        self.assertEqual(fields[7].kind, CursorKind.FIELD_DECL)
        self.assertIsNotNone(fields[7].translation_unit)
        self.assertEqual(fields[7].spelling, "h")
        self.assertFalse(fields[7].type.is_const_qualified())
        self.assertEqual(fields[7].type.kind, TypeKind.POINTER)
        self.assertEqual(fields[7].type.get_pointee().kind, TypeKind.POINTER)
        self.assertEqual(
            fields[7].type.get_pointee().get_pointee().kind, TypeKind.POINTER
        )
        self.assertEqual(
            fields[7].type.get_pointee().get_pointee().get_pointee().kind, TypeKind.INT
        )
        self.assertEqual(fields[7].type.get_typedef_name(), "")

    def test_references(self):
        """Ensure that a Type maintains a reference to a TranslationUnit."""

        tu = get_tu("int x;")
        children = list(tu.cursor.get_children())
        self.assertGreater(len(children), 0)

        cursor = children[0]
        t = cursor.type

        self.assertIsInstance(t.translation_unit, TranslationUnit)

        # Delete main TranslationUnit reference and force a GC.
        del tu
        gc.collect()
        self.assertIsInstance(t.translation_unit, TranslationUnit)

        # If the TU was destroyed, this should cause a segfault.
        t.get_declaration()

    def testConstantArray(self):
        tu = get_tu(CONSTARRAY_INPUT)

        teststruct = get_cursor(tu, "teststruct")
        self.assertIsNotNone(teststruct, "Didn't find teststruct??")
        fields = list(teststruct.get_children())
        self.assertEqual(fields[0].spelling, "A")
        self.assertEqual(fields[0].type.kind, TypeKind.CONSTANTARRAY)
        self.assertIsNotNone(fields[0].type.get_array_element_type())
        self.assertEqual(fields[0].type.get_array_element_type().kind, TypeKind.POINTER)
        self.assertEqual(fields[0].type.get_array_size(), 2)

    def test_equal(self):
        """Ensure equivalence operators work on Type."""
        source = "int a; int b; void *v;"
        tu = get_tu(source)

        a = get_cursor(tu, "a")
        b = get_cursor(tu, "b")
        v = get_cursor(tu, "v")

        self.assertIsNotNone(a)
        self.assertIsNotNone(b)
        self.assertIsNotNone(v)

        self.assertEqual(a.type, b.type)
        self.assertNotEqual(a.type, v.type)

        self.assertNotEqual(a.type, None)
        self.assertNotEqual(a.type, "foo")

    def test_type_spelling(self):
        """Ensure Type.spelling works."""
        tu = get_tu("int c[5]; void f(int i[]); int x; int v[x];")
        c = get_cursor(tu, "c")
        i = get_cursor(tu, "i")
        x = get_cursor(tu, "x")
        v = get_cursor(tu, "v")
        self.assertIsNotNone(c)
        self.assertIsNotNone(i)
        self.assertIsNotNone(x)
        self.assertIsNotNone(v)
        self.assertEqual(c.type.spelling, "int[5]")
        self.assertEqual(i.type.spelling, "int[]")
        self.assertEqual(x.type.spelling, "int")
        self.assertEqual(v.type.spelling, "int[x]")

    def test_typekind_spelling(self):
        """Ensure TypeKind.spelling works."""
        tu = get_tu("int a;")
        a = get_cursor(tu, "a")

        self.assertIsNotNone(a)
        self.assertEqual(a.type.kind.spelling, "Int")

    def test_function_argument_types(self):
        """Ensure that Type.argument_types() works as expected."""
        tu = get_tu("void f(int, int);")
        f = get_cursor(tu, "f")
        self.assertIsNotNone(f)

        args = f.type.argument_types()
        self.assertIsNotNone(args)
        self.assertEqual(len(args), 2)

        t0 = args[0]
        self.assertIsNotNone(t0)
        self.assertEqual(t0.kind, TypeKind.INT)

        t1 = args[1]
        self.assertIsNotNone(t1)
        self.assertEqual(t1.kind, TypeKind.INT)

        args2 = list(args)
        self.assertEqual(len(args2), 2)
        self.assertEqual(t0, args2[0])
        self.assertEqual(t1, args2[1])

    def test_argument_types_string_key(self):
        """Ensure that non-int keys raise a TypeError."""
        tu = get_tu("void f(int, int);")
        f = get_cursor(tu, "f")
        self.assertIsNotNone(f)

        args = f.type.argument_types()
        self.assertEqual(len(args), 2)

        with self.assertRaises(TypeError):
            args["foo"]

    def test_argument_types_negative_index(self):
        """Ensure that negative indexes on argument_types Raises an IndexError."""
        tu = get_tu("void f(int, int);")
        f = get_cursor(tu, "f")
        args = f.type.argument_types()

        with self.assertRaises(IndexError):
            args[-1]

    def test_argument_types_overflow_index(self):
        """Ensure that indexes beyond the length of Type.argument_types() raise."""
        tu = get_tu("void f(int, int);")
        f = get_cursor(tu, "f")
        args = f.type.argument_types()

        with self.assertRaises(IndexError):
            args[2]

    def test_argument_types_invalid_type(self):
        """Ensure that obtaining argument_types on a Type without them raises."""
        tu = get_tu("int i;")
        i = get_cursor(tu, "i")
        self.assertIsNotNone(i)

        with self.assertRaises(Exception):
            i.type.argument_types()

    def test_is_pod(self):
        """Ensure Type.is_pod() works."""
        tu = get_tu("int i; void f();")
        i = get_cursor(tu, "i")
        f = get_cursor(tu, "f")

        self.assertIsNotNone(i)
        self.assertIsNotNone(f)

        self.assertTrue(i.type.is_pod())
        self.assertFalse(f.type.is_pod())

    def test_function_variadic(self):
        """Ensure Type.is_function_variadic works."""

        source = """
#include <stdarg.h>

    void foo(int a, ...);
    void bar(int a, int b);
    """

        tu = get_tu(source)
        foo = get_cursor(tu, "foo")
        bar = get_cursor(tu, "bar")

        self.assertIsNotNone(foo)
        self.assertIsNotNone(bar)

        self.assertIsInstance(foo.type.is_function_variadic(), bool)
        self.assertTrue(foo.type.is_function_variadic())
        self.assertFalse(bar.type.is_function_variadic())

    def test_element_type(self):
        """Ensure Type.element_type works."""
        tu = get_tu("int c[5]; void f(int i[]); int x; int v[x];")
        c = get_cursor(tu, "c")
        i = get_cursor(tu, "i")
        v = get_cursor(tu, "v")
        self.assertIsNotNone(c)
        self.assertIsNotNone(i)
        self.assertIsNotNone(v)

        self.assertEqual(c.type.kind, TypeKind.CONSTANTARRAY)
        self.assertEqual(c.type.element_type.kind, TypeKind.INT)
        self.assertEqual(i.type.kind, TypeKind.INCOMPLETEARRAY)
        self.assertEqual(i.type.element_type.kind, TypeKind.INT)
        self.assertEqual(v.type.kind, TypeKind.VARIABLEARRAY)
        self.assertEqual(v.type.element_type.kind, TypeKind.INT)

    def test_invalid_element_type(self):
        """Ensure Type.element_type raises if type doesn't have elements."""
        tu = get_tu("int i;")
        ty = get_cursor(tu, "i").type
        with self.assertRaises(Exception) as ctx:
            ty.element_type
        self.assertEqual(str(ctx.exception), "Element type not available on this type.")

    def test_element_count(self):
        """Ensure Type.element_count works."""
        tu = get_tu("int i[5]; int j;")
        i = get_cursor(tu, "i")
        j = get_cursor(tu, "j")

        self.assertIsNotNone(i)
        self.assertIsNotNone(j)

        self.assertEqual(i.type.element_count, 5)

        with self.assertRaises(Exception):
            j.type.element_count

    def test_is_volatile_qualified(self):
        """Ensure Type.is_volatile_qualified works."""

        tu = get_tu("volatile int i = 4; int j = 2;")

        i = get_cursor(tu, "i")
        j = get_cursor(tu, "j")

        self.assertIsNotNone(i)
        self.assertIsNotNone(j)

        self.assertIsInstance(i.type.is_volatile_qualified(), bool)
        self.assertTrue(i.type.is_volatile_qualified())
        self.assertFalse(j.type.is_volatile_qualified())

    def test_is_restrict_qualified(self):
        """Ensure Type.is_restrict_qualified works."""

        tu = get_tu("struct s { void * restrict i; void * j; };")

        i = get_cursor(tu, "i")
        j = get_cursor(tu, "j")

        self.assertIsNotNone(i)
        self.assertIsNotNone(j)

        self.assertIsInstance(i.type.is_restrict_qualified(), bool)
        self.assertTrue(i.type.is_restrict_qualified())
        self.assertFalse(j.type.is_restrict_qualified())

    def test_get_result(self):
        tu = get_tu("void foo(); int bar(char, short);")
        foo = get_cursor(tu, "foo")
        bar = get_cursor(tu, "bar")
        self.assertEqual(foo.type.get_result().spelling, "void")
        self.assertEqual(bar.type.get_result().spelling, "int")

    def test_get_class_type(self):
        tu = get_tu(
            """
class myClass
{
   char *myAttr;
};

char *myClass::*pMyAttr = &myClass::myAttr;
""",
            lang="cpp",
        )
        cur = get_cursor(tu, "pMyAttr")
        self.assertEqual(cur.type.get_class_type().spelling, "myClass")

    def test_get_named_type(self):
        tu = get_tu("using char_alias = char; char_alias xyz;", lang="cpp")
        cur = get_cursor(tu, "xyz")
        self.assertEqual(cur.type.get_named_type().spelling, "char_alias")

    def test_get_ref_qualifier(self):
        tu = get_tu(
            """
class A
{
	const int& getAttr() const &;
	int getAttr() const &&;
};
""",
            lang="cpp",
        )
        getters = get_cursors(tu, "getAttr")
        self.assertEqual(len(getters), 2)
        self.assertEqual(getters[0].type.get_ref_qualifier(), RefQualifierKind.LVALUE)
        self.assertEqual(getters[1].type.get_ref_qualifier(), RefQualifierKind.RVALUE)

    def test_record_layout(self):
        """Ensure Cursor.type.get_size, Cursor.type.get_align and
        Cursor.type.get_offset works."""

        source = """
    struct a {
        long a1;
        long a2:3;
        long a3:4;
        long long a4;
    };
    """
        tries = [
            (["-target", "i386-linux-gnu"], (4, 16, 0, 32, 35, 64)),
            (["-target", "nvptx64-unknown-unknown"], (8, 24, 0, 64, 67, 128)),
            (["-target", "i386-pc-win32"], (8, 16, 0, 32, 35, 64)),
            (["-target", "msp430-none-none"], (2, 14, 0, 32, 35, 48)),
        ]
        for flags, values in tries:
            align, total, a1, a2, a3, a4 = values

            tu = get_tu(source, flags=flags)
            teststruct = get_cursor(tu, "a")
            fields = list(teststruct.get_children())

            self.assertEqual(teststruct.type.get_align(), align)
            self.assertEqual(teststruct.type.get_size(), total)
            self.assertEqual(teststruct.type.get_offset(fields[0].spelling), a1)
            self.assertEqual(teststruct.type.get_offset(fields[1].spelling), a2)
            self.assertEqual(teststruct.type.get_offset(fields[2].spelling), a3)
            self.assertEqual(teststruct.type.get_offset(fields[3].spelling), a4)
            self.assertEqual(fields[0].is_bitfield(), False)
            self.assertEqual(fields[1].is_bitfield(), True)
            self.assertEqual(fields[1].get_bitfield_width(), 3)
            self.assertEqual(fields[2].is_bitfield(), True)
            self.assertEqual(fields[2].get_bitfield_width(), 4)
            self.assertEqual(fields[3].is_bitfield(), False)

    def test_offset(self):
        """Ensure Cursor.get_record_field_offset works in anonymous records"""
        source = """
    struct Test {
      struct {int a;} typeanon;
      struct {
        int bariton;
        union {
          int foo;
        };
      };
      int bar;
    };"""
        tries = [
            (["-target", "i386-linux-gnu"], (4, 16, 0, 32, 64, 96)),
            (["-target", "nvptx64-unknown-unknown"], (8, 24, 0, 32, 64, 96)),
            (["-target", "i386-pc-win32"], (8, 16, 0, 32, 64, 96)),
            (["-target", "msp430-none-none"], (2, 14, 0, 32, 64, 96)),
        ]
        for _, values in tries:
            _, _, f1, bariton, foo, bar = values
            tu = get_tu(source)
            teststruct = get_cursor(tu, "Test")
            children = list(teststruct.get_children())
            fields = list(teststruct.type.get_fields())
            self.assertEqual(children[0].kind, CursorKind.STRUCT_DECL)
            self.assertNotEqual(children[0].spelling, "typeanon")
            self.assertEqual(children[1].spelling, "typeanon")
            self.assertEqual(fields[0].kind, CursorKind.FIELD_DECL)
            self.assertTrue(fields[0].is_anonymous())
            self.assertFalse(fields[0].is_anonymous_record_decl())
            self.assertEqual(fields[1].kind, CursorKind.FIELD_DECL)
            self.assertTrue(fields[1].is_anonymous())
            self.assertTrue(fields[1].is_anonymous_record_decl())
            self.assertEqual(teststruct.type.get_offset("typeanon"), f1)
            self.assertEqual(teststruct.type.get_offset("bariton"), bariton)
            self.assertEqual(teststruct.type.get_offset("foo"), foo)
            self.assertEqual(teststruct.type.get_offset("bar"), bar)

    def test_decay(self):
        """Ensure decayed types are handled as the original type"""

        tu = get_tu("void foo(int a[]);")
        foo = get_cursor(tu, "foo")
        a = foo.type.argument_types()[0]

        self.assertEqual(a.kind, TypeKind.INCOMPLETEARRAY)
        self.assertEqual(a.element_type.kind, TypeKind.INT)
        self.assertEqual(a.get_canonical().kind, TypeKind.INCOMPLETEARRAY)

    def test_addrspace(self):
        """Ensure the address space can be queried"""
        tu = get_tu("__attribute__((address_space(2))) int testInteger = 3;", "c")

        testInteger = get_cursor(tu, "testInteger")

        self.assertIsNotNone(testInteger, "Could not find testInteger.")
        self.assertEqual(testInteger.type.get_address_space(), 2)

    def test_template_arguments(self):
        source = """
        class Foo {
        };
        template <typename T>
        class Template {
        };
        Template<Foo> instance;
        int bar;
        """
        tu = get_tu(source, lang="cpp")

        # Varible with a template argument.
        cursor = get_cursor(tu, "instance")
        cursor_type = cursor.type
        self.assertEqual(cursor.kind, CursorKind.VAR_DECL)
        self.assertEqual(cursor_type.spelling, "Template<Foo>")
        self.assertEqual(cursor_type.get_num_template_arguments(), 1)
        template_type = cursor_type.get_template_argument_type(0)
        self.assertEqual(template_type.spelling, "Foo")

        # Variable without a template argument.
        cursor = get_cursor(tu, "bar")
        self.assertEqual(cursor.get_num_template_arguments(), -1)

    def test_pretty(self):
        tu = get_tu("struct X {}; X x;", lang="cpp")
        f = get_cursor(tu, "x")

        pp = PrintingPolicy.create(f)
        self.assertEqual(f.type.get_canonical().pretty_printed(pp), "X")
        pp.set_property(PrintingPolicyProperty.SuppressTagKeyword, False)
        self.assertEqual(f.type.get_canonical().pretty_printed(pp), "struct X")

    def test_fully_qualified_name(self):
        source = """
        namespace home {
          class Bar {
          };
          class Foo {
            public:
              void setIt(Bar*);
          };
        }
        class A : public home::Foo {
        };
        """
        tu = get_tu(source, lang="cpp")
        arg = next(get_cursor(tu, "setIt").get_arguments())
        pp = PrintingPolicy.create(arg)
        self.assertEqual(arg.type.get_fully_qualified_name(pp), "home::Bar *")
        self.assertEqual(arg.type.get_fully_qualified_name(pp, True), "::home::Bar *")

    def test_base_classes(self):
        source = """
        class A { int a; };
        class B { int b; };
        class C { int c; };
        template <typename T>
        class Template : public A, public B, virtual C {
        };
        Template<int> instance;
        int bar;
        """
        tu = get_tu(source, lang="cpp", flags=["--target=x86_64-linux-gnu"])
        cursor = get_cursor(tu, "instance")
        cursor_type = cursor.type
        cursor_type_decl = cursor_type.get_declaration()
        self.assertEqual(cursor.kind, CursorKind.VAR_DECL)
        bases = list(cursor_type.get_bases())
        self.assertEqual(len(bases), 3)
        self.assertFalse(bases[0].is_virtual_base())
        self.assertEqual(bases[0].get_base_offsetof(cursor_type_decl), 64)
        self.assertFalse(bases[1].is_virtual_base())
        self.assertEqual(bases[1].get_base_offsetof(cursor_type_decl), 96)
        self.assertTrue(bases[2].is_virtual_base())
        self.assertEqual(bases[2].get_base_offsetof(cursor_type_decl), 128)

    def test_class_methods(self):
        source = """
        template <typename T>
        class Template { void Foo(); };
        typedef Template<int> instance;
        instance bar;
        """
        tu = get_tu(source, lang="cpp", flags=["--target=x86_64-linux-gnu"])
        cursor = get_cursor(tu, "instance")
        cursor_type = cursor.underlying_typedef_type
        self.assertEqual(cursor.kind, CursorKind.TYPEDEF_DECL)
        methods = list(cursor_type.get_methods())
        self.assertEqual(len(methods), 4)
        self.assertEqual(methods[0].kind, CursorKind.CXX_METHOD)
        self.assertEqual(methods[1].kind, CursorKind.CONSTRUCTOR)
        self.assertEqual(methods[2].kind, CursorKind.CONSTRUCTOR)
        self.assertEqual(methods[3].kind, CursorKind.CONSTRUCTOR)



import os

from clang.cindex import Config, TokenKind

if "CLANG_LIBRARY_PATH" in os.environ:
    Config.set_library_path(os.environ["CLANG_LIBRARY_PATH"])

import unittest


class TestTokenKind(unittest.TestCase):
    def test_registration(self):
        """Ensure that items registered appear as class attributes."""
        self.assertTrue(hasattr(TokenKind, "LITERAL"))
        literal = TokenKind.LITERAL

        self.assertIsInstance(literal, TokenKind)

    def test_from_value(self):
        """Ensure registered values can be obtained from from_value()."""
        t = TokenKind.from_value(3)
        self.assertIsInstance(t, TokenKind)
        self.assertEqual(t, TokenKind.LITERAL)

    def test_repr(self):
        """Ensure repr() works."""

        r = repr(TokenKind.LITERAL)
        self.assertEqual(r, "TokenKind.LITERAL")



#!/usr/bin/env python
#
# ===- check-analyzer-fixit.py - Static Analyzer test helper ---*- python -*-===#
#
# Part of the LLVM Project, under the Apache License v2.0 with LLVM Exceptions.
# See https://llvm.org/LICENSE.txt for license information.
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
# ===------------------------------------------------------------------------===#
#
# This file copy-pasted mostly from the Clang-Tidy's 'check_clang_tidy.py'.
#
# ===------------------------------------------------------------------------===#

r"""
Clang Static Analyzer test helper
=================================

This script runs the Analyzer in fix-it mode and verify fixes, warnings, notes.

Usage:
  check-analyzer-fixit.py <source-file> <temp-file> [analyzer arguments]

Example:
  // RUN: %check-analyzer-fixit %s %t -analyzer-checker=core
"""

import argparse
import os
import re
import subprocess
import sys


def write_file(file_name, text):
    with open(file_name, "w") as f:
        f.write(text)


def run_test_once(args, extra_args):
    input_file_name = args.input_file_name
    temp_file_name = args.temp_file_name
    clang_analyzer_extra_args = extra_args

    file_name_with_extension = input_file_name
    _, extension = os.path.splitext(file_name_with_extension)
    if extension not in [".c", ".hpp", ".m", ".mm"]:
        extension = ".cpp"
    temp_file_name = temp_file_name + extension

    with open(input_file_name, "r") as input_file:
        input_text = input_file.read()

    # Remove the contents of the CHECK lines to avoid CHECKs matching on
    # themselves.  We need to keep the comments to preserve line numbers while
    # avoiding empty lines which could potentially trigger formatting-related
    # checks.
    cleaned_test = re.sub("// *CHECK-[A-Z0-9\-]*:[^\r\n]*", "//", input_text)
    write_file(temp_file_name, cleaned_test)

    original_file_name = temp_file_name + ".orig"
    write_file(original_file_name, cleaned_test)

    try:
        builtin_include_dir = subprocess.check_output(
            ["clang", "-print-file-name=include"], stderr=subprocess.STDOUT
        ).decode()
    except subprocess.CalledProcessError as e:
        print("Cannot print Clang include directory: " + e.output.decode())

    builtin_include_dir = os.path.normpath(builtin_include_dir)

    args = (
        [
            "clang",
            "-cc1",
            "-internal-isystem",
            builtin_include_dir,
            "-nostdsysteminc",
            "-analyze",
            "-analyzer-constraints=range",
            "-analyzer-config",
            "apply-fixits=true",
        ]
        + clang_analyzer_extra_args
        + ["-verify", temp_file_name]
    )

    print("Running " + str(args) + "...")

    try:
        clang_analyzer_output = subprocess.check_output(
            args, stderr=subprocess.STDOUT
        ).decode()
    except subprocess.CalledProcessError as e:
        print("Clang Static Analyzer test failed:\n" + e.output.decode())
        raise

    print(
        "----------------- Clang Static Analyzer output -----------------\n"
        + clang_analyzer_output
        + "\n--------------------------------------------------------------"
    )

    try:
        diff_output = subprocess.check_output(
            ["diff", "-u", original_file_name, temp_file_name], stderr=subprocess.STDOUT
        )
    except subprocess.CalledProcessError as e:
        diff_output = e.output

    print(
        "----------------------------- Fixes ----------------------------\n"
        + diff_output.decode()
        + "\n--------------------------------------------------------------"
    )

    try:
        subprocess.check_output(
            [
                "FileCheck",
                "-input-file=" + temp_file_name,
                input_file_name,
                "-check-prefixes=CHECK-FIXES",
                "-strict-whitespace",
            ],
            stderr=subprocess.STDOUT,
        )
    except subprocess.CalledProcessError as e:
        print("FileCheck failed:\n" + e.output.decode())
        raise


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("input_file_name")
    parser.add_argument("temp_file_name")

    args, extra_args = parser.parse_known_args()
    run_test_once(args, extra_args)


if __name__ == "__main__":
    main()



import os

from clang.cindex import CompilationDatabase, CompilationDatabaseError, Config

if "CLANG_LIBRARY_PATH" in os.environ:
    Config.set_library_path(os.environ["CLANG_LIBRARY_PATH"])

import gc
import unittest
import sys
from pathlib import Path

INPUTS_DIR = os.path.join(os.path.dirname(__file__), "INPUTS")


@unittest.skipIf(sys.platform == "win32", "TODO: Fix these tests on Windows")
class TestCDB(unittest.TestCase):
    def test_create_fail(self):
        """Check we fail loading a database with an assertion"""
        path = os.path.dirname(__file__)

        # clang_CompilationDatabase_fromDirectory calls fprintf(stderr, ...)
        # Suppress its output.
        try:
            stderr = os.dup(2)
            with open(os.devnull, "wb") as null:
                os.dup2(null.fileno(), 2)
            with self.assertRaises(CompilationDatabaseError) as cm:
                CompilationDatabase.fromDirectory(path)
        # Ensures that stderr is reset even if the above code crashes
        finally:
            os.dup2(stderr, 2)
            os.close(stderr)

        e = cm.exception
        self.assertEqual(e.cdb_error, CompilationDatabaseError.ERROR_CANNOTLOADDATABASE)

    def test_create(self):
        """Check we can load a compilation database"""
        CompilationDatabase.fromDirectory(INPUTS_DIR)

    def test_lookup_succeed(self):
        """Check we get some results if the file exists in the db"""
        cdb = CompilationDatabase.fromDirectory(INPUTS_DIR)
        cmds = cdb.getCompileCommands("/home/john.doe/MyProject/project.cpp")
        self.assertNotEqual(len(cmds), 0)

    def test_lookup_succeed_pathlike(self):
        """Same as test_lookup_succeed, but with PathLikes"""
        cdb = CompilationDatabase.fromDirectory(Path(INPUTS_DIR))
        cmds = cdb.getCompileCommands(Path("/home/john.doe/MyProject/project.cpp"))
        self.assertNotEqual(len(cmds), 0)

    def test_all_compilecommand(self):
        """Check we get all results from the db"""
        cdb = CompilationDatabase.fromDirectory(INPUTS_DIR)
        cmds = cdb.getAllCompileCommands()
        self.assertEqual(len(cmds), 3)
        expected = [
            {
                "wd": "/home/john.doe/MyProject",
                "file": "/home/john.doe/MyProject/project.cpp",
                "line": [
                    "clang++",
                    "--driver-mode=g++",
                    "-o",
                    "project.o",
                    "-c",
                    "/home/john.doe/MyProject/project.cpp",
                ],
            },
            {
                "wd": "/home/john.doe/MyProjectA",
                "file": "/home/john.doe/MyProject/project2.cpp",
                "line": [
                    "clang++",
                    "--driver-mode=g++",
                    "-o",
                    "project2.o",
                    "-c",
                    "/home/john.doe/MyProject/project2.cpp",
                ],
            },
            {
                "wd": "/home/john.doe/MyProjectB",
                "file": "/home/john.doe/MyProject/project2.cpp",
                "line": [
                    "clang++",
                    "--driver-mode=g++",
                    "-DFEATURE=1",
                    "-o",
                    "project2-feature.o",
                    "-c",
                    "/home/john.doe/MyProject/project2.cpp",
                ],
            },
        ]
        for i in range(len(cmds)):
            self.assertEqual(cmds[i].directory, expected[i]["wd"])
            self.assertEqual(cmds[i].filename, expected[i]["file"])
            for arg, exp in zip(cmds[i].arguments, expected[i]["line"]):
                self.assertEqual(arg, exp)

    def test_1_compilecommand(self):
        """Check file with single compile command"""
        cdb = CompilationDatabase.fromDirectory(INPUTS_DIR)
        file = "/home/john.doe/MyProject/project.cpp"
        cmds = cdb.getCompileCommands(file)
        self.assertEqual(len(cmds), 1)
        self.assertEqual(cmds[0].directory, os.path.dirname(file))
        self.assertEqual(cmds[0].filename, file)
        expected = [
            "clang++",
            "--driver-mode=g++",
            "-o",
            "project.o",
            "-c",
            "/home/john.doe/MyProject/project.cpp",
        ]
        for arg, exp in zip(cmds[0].arguments, expected):
            self.assertEqual(arg, exp)

    def test_2_compilecommand(self):
        """Check file with 2 compile commands"""
        cdb = CompilationDatabase.fromDirectory(INPUTS_DIR)
        cmds = cdb.getCompileCommands("/home/john.doe/MyProject/project2.cpp")
        self.assertEqual(len(cmds), 2)
        expected = [
            {
                "wd": "/home/john.doe/MyProjectA",
                "line": [
                    "clang++",
                    "--driver-mode=g++",
                    "-o",
                    "project2.o",
                    "-c",
                    "/home/john.doe/MyProject/project2.cpp",
                ],
            },
            {
                "wd": "/home/john.doe/MyProjectB",
                "line": [
                    "clang++",
                    "--driver-mode=g++",
                    "-DFEATURE=1",
                    "-o",
                    "project2-feature.o",
                    "-c",
                    "/home/john.doe/MyProject/project2.cpp",
                ],
            },
        ]
        for i in range(len(cmds)):
            self.assertEqual(cmds[i].directory, expected[i]["wd"])
            for arg, exp in zip(cmds[i].arguments, expected[i]["line"]):
                self.assertEqual(arg, exp)

    def test_compilecommand_iterator_stops(self):
        """Check that iterator stops after the correct number of elements"""
        cdb = CompilationDatabase.fromDirectory(INPUTS_DIR)
        count = 0
        for cmd in cdb.getCompileCommands("/home/john.doe/MyProject/project2.cpp"):
            count += 1
            self.assertLessEqual(count, 2)

    def test_compilationDB_references(self):
        """Ensure CompilationsCommands are independent of the database"""
        cdb = CompilationDatabase.fromDirectory(INPUTS_DIR)
        cmds = cdb.getCompileCommands("/home/john.doe/MyProject/project.cpp")
        del cdb
        gc.collect()
        cmds[0].directory

    def test_compilationCommands_references(self):
        """Ensure CompilationsCommand keeps a reference to CompilationCommands"""
        cdb = CompilationDatabase.fromDirectory(INPUTS_DIR)
        cmds = cdb.getCompileCommands("/home/john.doe/MyProject/project.cpp")
        del cdb
        cmd0 = cmds[0]
        del cmds
        gc.collect()
        cmd0.directory
